/**
 * Copyright 2005 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Id: WhitelistURLFilter.java 4485 2005-12-19 14:11:10Z kangas $
 */
package org.apache.nutch.net;

import org.apache.nutch.fs.*;
import org.apache.nutch.io.*;
import org.apache.nutch.net.URLFilter;
import org.apache.nutch.plugin.Extension;
import org.apache.nutch.plugin.PluginRepository;
import org.apache.nutch.util.LogFormatter;
import org.apache.nutch.util.NutchConf;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.regex.PatternSyntaxException;
import java.net.URL;
import java.net.MalformedURLException;

/** 
 * "White List" URLFilter.
 *
 * This filter accepts/rejects URLs based on a on-disk map file.
 * The map file is a hashtable that containes host/domain names as keys 
 * and regexes as values. Host and/or domain names are extracted from the URL 
 * and used as keys to lookup into the hash table.
 * The full description of the algorithm is explained in the filter() method.
 *
 * We use the Java 1.4 regex engine (java.util.regex)
 */
public class WhitelistURLFilter implements URLFilter {
  private static final Logger LOG = LogFormatter.getLogger(WhitelistURLFilter.class.getName());
  private static final String LOGPREFIX = "WhitelistFilter: ";
  private static final String PLUGIN_NAME = "urlfilter-whitelist";
  private static final String whitelistFileProp = "urlfilter.whitelist.file";
  private static final String undirectedCrawlProp = "urlfilter.whitelist.enableUndirectedCrawl";

  private NutchFileSystem nfs;
  private MapFile.Reader whitelistMap;
  private boolean undirectedCrawl;

  /**
   * Constructor that does not read NutchConf
   */
  public WhitelistURLFilter(String mapFileDir, boolean undirectedCrawl) 
  throws IOException {
    init(mapFileDir, undirectedCrawl);
  }
  
  /**
   * Constructor that reads NutchConf to discover:
   * - path to the MapFile directory
   * - whether to do an undirected crawl or not
   */
  public WhitelistURLFilter() throws IOException {
    // Read the location of the MapFile
    NutchConf nutchConf = NutchConf.get();
    boolean undirectedCrawl = nutchConf.getBoolean(undirectedCrawlProp, false);
    String mapFileDir = nutchConf.get(whitelistFileProp);
    if (mapFileDir == null) {
      throw new IOException(LOGPREFIX +"can't find config property \"" 
			    + whitelistFileProp + "\"!");
    }
    
    LOG.fine(LOGPREFIX +"mapFileDir:"+ mapFileDir);
    LOG.fine(LOGPREFIX +"undirectedCrawl ? "+ undirectedCrawl);
    init(mapFileDir, undirectedCrawl);
  }
  
  /* used by both constructors */
  private void init(String mapFileDir, boolean undirectedCrawl) 
  throws IOException {
    // Open the map file for future lookups in filter()
    this.nfs = new LocalFileSystem();
    try {
      whitelistMap = new MapFile.Reader(this.nfs, mapFileDir);
    } catch (IOException e) {
      LOG.warning(LOGPREFIX+"can't read MapFile: "+ mapFileDir);
      throw e;
    }
  }


  /**
   * Split the regexes string into individual regexes 
   * and match them against the stripped URL.
   * @return true if the input string matches any of the regexes.
   */
   private boolean matchRegexes(String strippedURL, String regexesTxt) {
    
    if (strippedURL == null) return false;
    String[] regexes = regexesTxt.split("\\s");
    
    for (int i=0; i < regexes.length; i++) {
      String re = regexes[i];
      if (re == null) continue;
      
      try {
	if (strippedURL.matches(re)) {
	  LOG.fine(LOGPREFIX +"match found: \"" + strippedURL + "\" ~= \"" + re + "\"");
	  return true;
	}
      } catch (PatternSyntaxException pe) {
	LOG.log(Level.WARNING, LOGPREFIX +"regex parsing error", pe);
      }
    }
    
    return false;
  }
  
  /* Tiny wrapper to provide debug output */
  private String whitelistMatch(String url, String strippedURL, String regexes) {
    if (matchRegexes(strippedURL, regexes)) {
      LOG.fine(LOGPREFIX +"kept url (match): " + url);
      return url;
    } else {
      LOG.fine(LOGPREFIX +"rejected url (no match): " + url);
      return null;
    }
  }
  
  /**
   * The whitelist filter works as follow:
   * Look up hostname in map file
   * if found, then
   *   iterate through list of regexes and return match status 
   * else
   *   trim off first dot in hostname and look up in map file
   *   if found, then
   *     iterate through list of regexes and return match status
   *   else
   *     if undirected crawl is set
   *       test if URL contains "CGI" character (one of [?*!@=])
   *       if yes, then
   *         return null
   *       else
   *         return string 
   *     else
   *       return null
   */
  public synchronized String filter(String url) {
    LOG.finer(LOGPREFIX + "url = \"" + url + "\"");
    
    boolean isCgiUrl = false;
    String strippedURL;
    String hostname;
    
    try {
      URL u = new URL(url);
      hostname = u.getHost();
      isCgiUrl = u.getQuery() != null;
      strippedURL = u.getFile();
    } 
    catch(MalformedURLException me) {
      LOG.warning(LOGPREFIX + "MalformedURLException url: " + url);
      return null;
    }
    
    // if we're in undirectedCrawl, we always OK non-CGI URLs
    if (undirectedCrawl && !isCgiUrl) {
      LOG.fine(LOGPREFIX + "kept url (undirected): " + url);
      return url;
    }

    // Look first for the hostname, then domain, in the MapFile
    UTF8 value = new UTF8();
    try {
      LOG.finest(LOGPREFIX + "extracting hostname: \"" + hostname + "\"");
      UTF8 entry = (UTF8) whitelistMap.get(new UTF8(hostname), value);
      
      // look for domain?
      if (entry == null) {
        String domain = extractDomainFromHostname(hostname);
        LOG.finer(LOGPREFIX + "hostname entry not found in map file");
        LOG.finest(LOGPREFIX + "extracing domain: \"" + domain + "\"");
        entry = (UTF8) whitelistMap.get(new UTF8(domain), value);
      }
      
      // test regexes
      if (entry != null) {
        String regexes = entry.toString();
        LOG.finer(LOGPREFIX + "found entry in map file");
        return whitelistMatch(url, strippedURL, regexes);
      }
      LOG.finer(LOGPREFIX + "no match found in the MapFile");
    } 
    catch(IOException e) {
      LOG.warning(LOGPREFIX + "map file read error: " + e);
    }

    // no match found.
    String msg = undirectedCrawl ? "undirected, CGI" : "directed";
    LOG.fine(LOGPREFIX + "rejected url ("+msg+"): " + url);
    return null;
  }
  
  /* Extract the domain name from a host name */
  private static String extractDomainFromHostname(String hostname) {
    String parts[] = hostname.split("\\.", 2);
    if (parts.length != 2) {
      LOG.warning(LOGPREFIX + ".extractDomainFromHostname: invalid hostname, '.' not found! \"" + hostname + "\"");
      return hostname;
    }
    return parts[1];
  }

  private static void usage() {
    System.out.println("Usage: WhitelistURLFilter <path-to-mapfile>");
    System.out.println("Reads: URLs from stdin");
    System.out.println("Prints +URL or -URL to stdout");
  }
  
  /**
   * main() for testing...
   */
  public static void main(String args[]) throws IOException {
    if (args.length != 1) {
      usage(); return;
    }
    
    WhitelistURLFilter filter = new WhitelistURLFilter(args[0], true);
    
    BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
    while (true) {
      String line = in.readLine();
      if (line == null) break;
      
      if (null != filter.filter(line)) {
	System.out.println("+"+line);
      } else {
	System.out.println("-"+line);
      }
    }
    in.close();
  }
}
