/**
 * Copyright 2005 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Id: WhitelistWriter.java 4480 2005-12-19 05:08:28Z kangas $
 */
package org.apache.nutch.net;

import org.apache.nutch.io.*;
import org.apache.nutch.fs.LocalFileSystem;
import org.apache.nutch.fs.NutchFileSystem;
import java.io.*;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Logger;
import org.apache.nutch.util.LogFormatter;

/**
 * This class writes whitelist expressions into a Nutch MapFile
 * There are 2 writeMapFile() methods, one that reads an array of strings 
 * as input and another that reads the input from a file.
 * 
 * The syntax of an input line is the following:
 * hostname=regex1 regex2 regex3 ...
 *
 * Examples:
 * www.bryantpark.org=/calendar/events/.+
 * petescandystore.com=/newmusic/newshows.html
 * www.playbill.com=/events/event_detail/.+ /events/3/.+
 */
public class WhitelistWriter {
  public static final Logger LOG = LogFormatter.getLogger(WhitelistWriter.class.getName());
  private static final String LOGPREFIX = "WhitelistWriter: ";
  private static final String TEMPPREFIX = "whitelistwriter";
  private static final Class KEY_CLASS = UTF8.class;
  private static final Class VAL_CLASS = UTF8.class;
  
  private NutchFileSystem nfs;
  private File mapFileDir;
  private boolean overwrite;
  private File tempFile;
  private SequenceFile.Writer tempWriter;
  
  /** Constructor: create tempFile and tempWriter */
  public WhitelistWriter(NutchFileSystem nfs, String dirName, boolean overwrite) 
  throws IOException {
    this.nfs = nfs;
    this.overwrite = overwrite;
    
    mapFileDir = new File(dirName);
    
    tempFile = File.createTempFile(TEMPPREFIX, null);
    tempFile.delete();		// want just a path
    tempWriter = new SequenceFile.Writer(nfs, tempFile.getAbsolutePath(),
    					 KEY_CLASS, VAL_CLASS);
  }
  
  
  /** Open the map file for writing */
  private MapFile.Writer openMapFile(String dirName, boolean overwrite) 
  throws IOException {
    if (overwrite) {
      try { MapFile.delete(nfs, dirName); } 
      catch (IOException e) {
        LOG.warning(LOGPREFIX +"error, can't overwrite the map file: " + dirName);
	throw e;
      }
    }
    try {
      return new MapFile.Writer(nfs, dirName, UTF8.class, UTF8.class);
    } 
    catch (IOException e) {
      LOG.warning(LOGPREFIX +"error, can't open the map file for writing: "+ dirName);
      throw e;
    }
  }
  
  /** 
   * Write one MapFile entry to the temporary SequenceFile, which we will sort.
   * @param key String
   * @param value String
   */
  public void append(String key, String value) throws IOException {
    tempWriter.append(new UTF8(key), new UTF8(value));
  }
  
  /** 
   * Sort temporary SequenceFile and copy data to MapFile
   * @return boolean : true if MapFile written, false if previously closed
   */
  public boolean close() throws IOException {
    if (tempWriter == null) return false;
    
    tempWriter.close();
    tempWriter = null;
    
    // sort tempfile
    File output = File.createTempFile(TEMPPREFIX+"-sort", null);
    output.delete();		// want just a path
    String outputPath = output.getAbsolutePath();
    SequenceFile.Sorter sorter = new SequenceFile.Sorter(nfs, KEY_CLASS, VAL_CLASS);
    
    sorter.sort(tempFile.getAbsolutePath(), outputPath);
    
    // we're now done with tempFile
    tempFile.delete();
    
    // now create MapFile and copy from sorted SequenceFile
    MapFile.Writer mapwriter = openMapFile(mapFileDir.getAbsolutePath(), overwrite);
    SequenceFile.Reader reader = new SequenceFile.Reader(nfs, outputPath);
    
    UTF8 key = new UTF8();
    UTF8 value = new UTF8();
    
    while (reader.next(key, value)) {
      mapwriter.append(key, value);
    }
    
    mapwriter.close();
    reader.close();
    nfs.close();
    
    output.delete();
    return true;
  }
  
  /**
   * Parse an input string into 2 substrings: host name, regular expressions
   * (see comments at beginning of file for more info about the input format)
   */
   private static String[] parseLine(String s) {
     String[] parts = s.split("=", 2);
     if (parts == null || parts.length != 2) {
       return null;
     }
     parts[1] = parts[1].trim();
     return parts;
   }

  /** 
   * Static convenience method. Used by TestWhitelistWriter.
   * We assume whitelistData is a List of Strings like "key=value value2"
   */
  public static boolean writeMapFile(List whitelistData, String dirName, boolean overwrite)
  throws IOException {
    NutchFileSystem nfs = new LocalFileSystem();
    WhitelistWriter w = new WhitelistWriter(nfs, dirName, overwrite);
    Iterator i = whitelistData.iterator();
    
    while (i.hasNext()) {
      String line = (String) i.next();
      String[] parts = parseLine(line);
      if (parts != null) { w.append(parts[0], parts[1]); }
    }
    return w.close();
  }
  
  /** print usage message */
  private static void usage() {
    System.out.println("Usage: WhitelistWriter <dir>");
    System.out.println("<dir>: a writeable directory name");
    System.out.println("Reads: whitelist expressions from stdin");
    System.out.println("Syntax:");
    System.out.println("\thostname=regex1 regex2 regex3");
    System.out.println("Examples:");
    System.out.println("\twww.bryantpark.org=/calendar/events/.+");
    System.out.println("\tpetescandystore.com=/newmusic/newshows.html");
    System.out.println("\twww.playbill.com=/events/event_detail/.+ /events/3/.+");
  }

  /** main method */
  public static void doMain(String[] args) throws IOException {
    if (args.length != 1) {
      usage(); return;
    }
    String mapfileDir = args[0];
    
    BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
    int count = 0;
    
    NutchFileSystem nfs = new LocalFileSystem();
    WhitelistWriter w = new WhitelistWriter(nfs, mapfileDir, true);
    
    while (true) {
      String line = in.readLine();
      if (line == null) 
	break;
      String[] parts = parseLine(line);
      if (parts != null) { 
	w.append(parts[0], parts[1]);
	count++;
      }
    }
    in.close();
    w.close();
    System.out.println("Total "+ count +" records written to "+mapfileDir);
  }
  
  /** main wrapper to set correct shell exit status */
  public static void main(String[] args) {
    try {
      doMain(args);
    } catch (Exception e) {
      e.printStackTrace();
      Runtime.getRuntime().exit(1);
    }
  }
}
